"""Helper functions for estout package"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_utils.ipynb.

# %% ../nbs/01_utils.ipynb 3
from __future__ import annotations
import os, platform, functools
from pathlib import Path
from subprocess import PIPE, Popen, run, SubprocessError
from typing import Dict, List, Tuple, Literal 
import pandas as pd 
import numpy as np

# %% auto 0
__all__ = ['rgetattr', 'rsetattr', 'default_formats', 'get_stars', 'model_groups', 'tex_table_env', 'df_to_tex',
           'make_pdf_from_tex', 'open_pdf_file']

# %% ../nbs/01_utils.ipynb 4
def rgetattr(obj, attr, *args):
    """Recursive getattr (for nested attributes)."""
    def _getattr(obj, attr):
        return getattr(obj, attr, *args)
    return functools.reduce(_getattr, [obj] + attr.split('.'))

# %% ../nbs/01_utils.ipynb 5
def rsetattr(obj, attr, val):
    """Recursive setattr (for nested attributes)."""
    pre, _, post = attr.rpartition('.')
    return setattr(rgetattr(obj, pre) if pre else obj, post, val)

# %% ../nbs/01_utils.ipynb 6
def default_formats() -> dict:
    """Default output formats for some very common statistics."""
    return  {'params':'{:.2f}', 
            'tstats':'{:.2f}', 
            'pvalues': '{:.3f}',
            'se':'{:.2f}', 
            'r2':'{:.3f}',
            'nobs':'{:.0f}'
            }

# %% ../nbs/01_utils.ipynb 8
def get_stars(pvalues: pd.Series, # This is compared to key of `stars` param to determine how many stars should be added
            stars: dict = {.1:'*',.05:'**',.01:'***'} # Todo: default values to the left are star symbols that are not rendered correctly in markdown
            ) -> pd.Series:
    """For each pvalue, check the lowest key in `stars` for which the pvalue is smaller than that key, and return the corresponding nr of stars."""

    param_names = list(pvalues.index)

    #Sort 'stars' by key (in reverse order)
    ks = list(stars.keys())
    ks.sort(reverse=True)
    stars = {k: stars[k] for k in ks}

    out = pd.Series('', index=param_names)
    for param in param_names:    
        for alpha in stars:
            if pvalues[param] < alpha:
                out[param] = stars[alpha]
                
    return out

# %% ../nbs/01_utils.ipynb 11
def model_groups(column_group_names: Dict[str, List[int]], # Keys are group titles, values are lists of column indices included in each group
                add_clines: bool=True # If True, adds lines below group names
                ) -> str:
    """Returns LaTex code needed to add at the top of the table in order to give names to groups of columns in the table."""
    
    group_names = ''
    group_lines = ''
    for key,value in column_group_names.items():
        if type(key) != str:
            raise TypeError('Each key in column_group_names must be a string')
        if type(value) != list:
            raise TypeError("Each value in column_group_names dict must be a list")
        if len(value) != 2:
            raise TypeError("Each value in column_group_names dict must contain two integers")
        
        value = sorted(value)
        group_names += '& \\multicolumn{%i}{c}{%s} ' %(value[1]-value[0]+1,key)
        if add_clines:
            group_lines += '\\cline{%s-%s} ' %(str(value[0]+1),str(value[1]+1))

    return group_names + ' \\\\ \n' + group_lines + ' \n'

# %% ../nbs/01_utils.ipynb 13
def tex_table_env(nr_columns: int, # Number of columns in the table
                    env: Literal['tabularx','tabular*']='tabular*' # Latex tabular environment specification
                    ) -> Tuple[str,str]:
    """Creates LaTex code to add at the top of the table to create the correct tabular environment."""

    if env=='tabularx':
        header = '\\begin{tabularx}{\\textwidth}{@{}l *{%i}{>{\\centering\\arraybackslash}X}@{}}' %nr_columns 
        footer = '\\end{tabularx}'
    elif env=='tabular*':
        header = '\\begin{tabular*}{\\textwidth}{@{\\extracolsep{\\fill}}l*{%i}{c}}' %nr_columns
        footer = '\\end{tabular*}'
    else:
        raise NotImplemented(f"LaTex tabular environment {env} has not yet been implemented in tex_table_env()")
    return header,footer 

# %% ../nbs/01_utils.ipynb 16
def df_to_tex(df: pd.DataFrame, # If this has a MultiIndex, only the first level is kept
                panel_title: str='',
                palign: Literal['l','r','c']='l', # Alignment of panel title 
                col_groups: Dict[str, List[int]]=None, # Keys are group names; values are lists of consecutive indices of columns in the group
                col_names: List[str]|bool=True, # If False, none; if True, use df column names; if list, gives custom column names
                hlines: List[int]=[], # Row indices under which to place hline (use 0 for topline). adds \smallskip under each line
                tabular_env: str='tabular*' #LaTex tabular environment
                ) -> str: 
    """Creates LaTeX-formatted table from DataFrame."""

    nr_cols = len(df.columns)
    df = df.astype('string')
    if isinstance(df.index, pd.MultiIndex): df = df.droplevel(1)

    header,footer = tex_table_env(nr_cols, tabular_env)
    out = header + ' \n'

    if panel_title:
        out += f'\\multicolumn{{{nr_cols+1}}}{{@{{}} {palign}}}{{{panel_title}}}' + ' \\\\ \n'

    if col_groups: out += model_groups(col_groups)
    if col_names is True: col_names = [str(x) for x in df.columns]
    if col_names: out += ' & '.join([''] + col_names) + ' \\\\ \n'
    for rownr in range(df.shape[0]):
        out += str(df.index[rownr]) + ' & '  + ' & '.join(list(df.iloc[rownr])) +  ' \\\\ \n' 
    out = out + footer

    linebreaks = [i for i in range(len(out)) if out[i:].startswith('\n')]
    for line_nr, pos in reversed(list(enumerate(linebreaks))):
        if line_nr in hlines: out = out[:pos+1] + ' \\hline \\noalign{\\smallskip} \n' + out[pos+1:]

    return out

# %% ../nbs/01_utils.ipynb 20
def make_pdf_from_tex(tex_file_path: Path|str) -> Path:
    """Output PDF is created in the same folder as source tex file. Requires TexLive and its pdflatex utility"""

    if isinstance(tex_file_path,str): tex_file_path = Path(tex_file_path)
    os.chdir(tex_file_path.parent)   
    process = Popen(['pdflatex',tex_file_path], stdin=PIPE, stdout=PIPE, stderr=PIPE,shell=False)
    _ , errors = process.communicate()
    if process.returncode == 0: print("PDF creation successful!")
    else: print("PDF creation failed. Errors:", errors.decode('utf-8'))
    return Path(str.replace(str(tex_file_path), '.tex', '.pdf'))

# %% ../nbs/01_utils.ipynb 21
def open_pdf_file(file_path):
    try:
        if platform.system() == "Windows": run(['start', file_path])
        else: run(['open', file_path])
    except FileNotFoundError:
        print(f"File not found: {file_path}")
    except SubprocessError:
        print(f"Error opening PDF file: {file_path}")



"""Functions that extract and standardize results from `linearmodels.PanelOLS.fit()' objects."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_linearmodels_results.ipynb.

# %% auto 0
__all__ = ['ynames', 'xnames', 'params', 'tstats', 'pvalues', 'covmat', 'se', 'nobs', 'r2']

# %% ../nbs/03_linearmodels_results.ipynb 4
import numpy as np
import pandas as pd

# %% ../nbs/03_linearmodels_results.ipynb 8
def ynames(res): return res.model.dependent.vars 

# %% ../nbs/03_linearmodels_results.ipynb 10
def xnames(res): return res.model.exog.vars

# %% ../nbs/03_linearmodels_results.ipynb 12
def params(res): return res.params

# %% ../nbs/03_linearmodels_results.ipynb 15
def tstats(res): return res.tstats

# %% ../nbs/03_linearmodels_results.ipynb 17
def pvalues(res): return res.pvalues

# %% ../nbs/03_linearmodels_results.ipynb 19
def covmat(res): return res.cov

# %% ../nbs/03_linearmodels_results.ipynb 21
def se(res): return res.std_errors

# %% ../nbs/03_linearmodels_results.ipynb 23
def nobs(res): return res.nobs

# %% ../nbs/03_linearmodels_results.ipynb 25
def r2(res): return res.rsquared



"""Main functions of `estout` package"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_core.ipynb.

# %% ../nbs/00_core.ipynb 4
from __future__ import annotations
from pathlib import Path 
from typing import List, Dict, Literal, Union, Callable
import importlib

import numpy as np
import pandas as pd
import statsmodels.api as sm
from linearmodels import PanelOLS

from . import utils 

# %% auto 0
__all__ = ['collect_stats', 'to_df', 'to_tex', 'to_pdf']

# %% ../nbs/00_core.ipynb 7
def collect_stats(res, # Results object to extract stats from
                  get_default_stats = True, # If True, returns all stats implemented by the `f'{package}_results'` module
                  add_stats: dict=None, # Keys are stats to extract in addition to the default ones; values are attributes of `res` or callables
                  add_literals: dict=None, # Additional info to be added to output dict; values must be scalars
                  ) -> dict:
    """Collects stats from `res` object. Stats in `add_stats` can override default stats"""

    out = {'package': res.__module__.split('.')[0]}
    results_module = importlib.import_module(f"estout.{out['package']}_results")

    if get_default_stats:
        for stat in results_module.__all__:
            out[stat] = utils.rgetattr(results_module, stat)(res)

    if add_stats is not None:
        for stat, val in add_stats.items():
            if isinstance(val, str): out[stat] = utils.rgetattr(res, val)
            elif isinstance(val, Callable): out[stat] = val(res)
            else: raise TypeError("Values of add_stats dict must be either strings or callables")

    if add_literals is not None:
        out.update(add_literals)
        
    return out

# %% ../nbs/00_core.ipynb 13
def to_df(res_list: List[dict], # List of outputs from `collect_stats()`
          which_xvars: list=None, # If None, report all xvars
          stats_body: list=['params', 'tstats'], # Each element of `res_list` needs to have these stats as keys; values must be pd.Series
          stats_bottom: list=['r2', 'nobs'], # Each element of `res_list` needs to have these stats as keys; values must be scalars
          labels: dict=None,
          add_formats: dict=None  
          ) -> pd.DataFrame: 
    """Combines results from multiple `collect_stats()` outputs into a single `pd`.DataFrame"""  
    
    formats = utils.default_formats()
    if add_formats is not None: formats.update(add_formats)
    
    columns = []
    for i,res in enumerate(res_list):
        newcol = pd.concat([res[x] for x in stats_body], axis=1, ignore_index=True).set_axis(stats_body, axis=1)
        for x in stats_body:
            newcol[x] = newcol[x].map(formats[x].format)
            if x == 'params':
                newcol[x] += utils.get_stars(res['pvalues'])
            else:
                newcol[x] = '(' + newcol[x] + ')'
        newcol = newcol.stack(level=0) #set_index('coeff_names')
        columns.append(newcol)

    out = pd.concat(columns, axis = 1)
    if which_xvars is not None: out = out.loc[which_xvars].copy()
    
    for i,res in enumerate(res_list):
        for x in stats_bottom:
            out.loc[x,i] = formats[x].format(res[x]) if x in formats else res[x]

    if labels is not None:
        for var in set(out.droplevel(1).index):
            if var in labels: out = out.rename(index={var:labels[var]}, level=0) 

    #Clean up row names 
    out = out.reset_index()
    out.loc[~out['level_1'].isin(['params','']), 'level_0'] = ''
    out = out.set_index(['level_0','level_1'])
    out.index.names = (None,None)

    return out.astype('string').fillna('')

# %% ../nbs/00_core.ipynb 16
def to_tex(dfs: pd.DataFrame|List[pd.DataFrame], # DataFrame(s) to be converted to tex table; if multiple, they will be panels in a larger table
            outfile: Path|str=None, # Where to save resulting tex output
            title: str='Table title', # Table title
            notes: str='Table description', # Some call this the table caption
            notes_on_top: bool=True, # Set to False if you want table description (caption) to be at the bottom
            label: str='', # Table label (for referencing within LaTex document)
            table_type: Literal['table','sidewaystable']='table',
            font_size: str='\\footnotesize', # Gets applied to the table contents as well as its caption
            addtocounter: int=0, # Set to -1 for tables that are just a continuation of a table on a new page
            
            panel_title: List[str]=None, # One element in the list for each dataframe in `dfs`
            palign: Literal['l','r','c']='l', # Alignment of panel title 
            col_groups: List[dict]=None, # Keys are group names; values are lists of consecutive indices of columns in the group
            col_names: List[Union[list,bool]]=True, # If False, none; if True, use df column names; if list, gives custom column names
            hlines: List[List[int]]=None, # Row indices under which to place hline
            tabular_env: str='tabular*' #LaTex tabular environment
            ) -> str:
    """Create tex code to generate table from one or more dataframes"""

    if isinstance(dfs, pd.DataFrame): dfs = [dfs]
    if panel_title is None: panel_title = ['']*len(dfs)
    if col_groups is None: col_groups = [None]*len(dfs)
    if col_names in [True, False]: col_names = [col_names]*len(dfs)
    if hlines is None: hlines = [[]]*len(dfs)
    body =  '\n \\smallskip \n'.join([utils.df_to_tex(dfs[i], panel_title=panel_title[i], 
                                                        palign=palign, tabular_env=tabular_env,
                                                        col_groups=col_groups[i], col_names=col_names[i], hlines=hlines[i]) 
                            for i in range(len(dfs))])
    
    pre = "\\newpage \n \\clearpage \n "
    pre += f"\\begin{{{table_type}}}[!h] {font_size} \n"
    pre += f"\\addtocounter{{table}}{{{addtocounter}}} \n"
    pre += f"\\caption{{\\textbf{{{title}}}}} \n"

    post = f"\\label{{{label}}} \n \\end{{{table_type}}} \n"

    notes_tex = ''
    if addtocounter==0: notes_tex = f"\\par {{{notes}}}"

    if notes_on_top:
        mid = notes_tex + f" \n\n \\vspace{{{'2mm'}}} \n\n {body} \n"
    else:
        mid = f"{body} \n\n \\vspace{{{'2mm'}}} \n\n " + notes_tex + ' \n'

    content = pre + mid + post

    if outfile:
        with open(outfile, "+w") as f:
            f.write(content) 
    return content

# %% ../nbs/00_core.ipynb 18
def to_pdf(outfile: str, # Path to .tex file where combined tables are saved (must contain .tex extension)
            table_tex_code: str|Path|List[Union[str,Path]]=None, # String(s) or Paths to files containing table tex code (e.g. like outputs of to_tex())
            article_spec=r'\documentclass[11pt]{article}',
            captionsetup="format=plain, labelsep=newline, labelfont = bf, justification=centering",
            make_pdf: bool=True, 
            open_pdf: bool=False):
    r"""Creates PDF with one or more tables given their tex code (from \being{table} to \end{table})"""

    if isinstance(table_tex_code, str): table_tex_code = [table_tex_code]
    tex_strings = []
    for t in table_tex_code:
        if isinstance(t, Path):
            with open(t,'r') as f:
                tex_strings.append(f.read())
        else: tex_strings.append(t)
    tables = ' \n '.join(tex_strings)

    preamble = r"\usepackage{booktabs,setspace,graphicx,epstopdf,tabularx, bigstrut,textcomp, outlines}" + "\n"
    preamble += r"\usepackage{amsmath,amsfonts,amssymb,amsthm,caption}" + "\n"
    preamble += r"\usepackage[margin=1in]{geometry}" + "\n"
    preamble += f"\\captionsetup{{{captionsetup}}}" + "\n"    

    content = "\n".join([article_spec, preamble, r'\begin{document}', tables, r'\end{document}'])

    with open(outfile, "w") as f:
        f.write(content) 
    if make_pdf:
        pdf_path = utils.make_pdf_from_tex(outfile)
        if open_pdf: return utils.open_pdf_file(pdf_path)



"""Functions that extract and standardize results from `statsmodels.api.OLS.fit()` objects and `statsmodels.api.OLS.fit().get_robustcov_results()` objects."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_statsmodels_results.ipynb.

# %% auto 0
__all__ = ['ynames', 'xnames', 'params', 'tstats', 'pvalues', 'covmat', 'se', 'nobs', 'r2']

# %% ../nbs/02_statsmodels_results.ipynb 4
import numpy as np
import pandas as pd

# %% ../nbs/02_statsmodels_results.ipynb 9
def ynames(res): return list(res.model.endog_names)

# %% ../nbs/02_statsmodels_results.ipynb 11
def xnames(res): return res.model.exog_names

# %% ../nbs/02_statsmodels_results.ipynb 13
def params(res): 
    if isinstance(res.params, np.ndarray):
            return pd.Series(res.params, index=xnames(res)) 
    else:
        return res.params

# %% ../nbs/02_statsmodels_results.ipynb 16
def tstats(res):
    if isinstance(res.tvalues, np.ndarray):
            return pd.Series(res.tvalues, index=xnames(res)) 
    else:
        return res.tvalues

# %% ../nbs/02_statsmodels_results.ipynb 19
def pvalues(res):
    if isinstance(res.pvalues, np.ndarray):
            return pd.Series(res.pvalues, index=xnames(res)) 
    else:
        return res.pvalues

# %% ../nbs/02_statsmodels_results.ipynb 22
def covmat(res):
    if isinstance(res.cov_params(), np.ndarray):
            return pd.DataFrame(res.cov_params(), index=xnames(res), columns=xnames(res)) 
    else:
        return res.cov_params()

# %% ../nbs/02_statsmodels_results.ipynb 25
def se(res): return pd.Series(np.sqrt(np.diag(np.array(covmat(res)))),index=xnames(res))

# %% ../nbs/02_statsmodels_results.ipynb 28
def nobs(res): return int(res.nobs)

# %% ../nbs/02_statsmodels_results.ipynb 30
def r2(res): return res.rsquared



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/estout',
                'doc_host': 'https://ionmihai.github.io',
                'git_url': 'https://github.com/ionmihai/estout',
                'lib_path': 'estout'},
  'syms': { 'estout.core': { 'estout.core.collect_stats': ('core.html#collect_stats', 'estout/core.py'),
                             'estout.core.to_df': ('core.html#to_df', 'estout/core.py'),
                             'estout.core.to_pdf': ('core.html#to_pdf', 'estout/core.py'),
                             'estout.core.to_tex': ('core.html#to_tex', 'estout/core.py')},
            'estout.linearmodels_results': { 'estout.linearmodels_results.covmat': ( 'linearmodels_results.html#covmat',
                                                                                     'estout/linearmodels_results.py'),
                                             'estout.linearmodels_results.nobs': ( 'linearmodels_results.html#nobs',
                                                                                   'estout/linearmodels_results.py'),
                                             'estout.linearmodels_results.params': ( 'linearmodels_results.html#params',
                                                                                     'estout/linearmodels_results.py'),
                                             'estout.linearmodels_results.pvalues': ( 'linearmodels_results.html#pvalues',
                                                                                      'estout/linearmodels_results.py'),
                                             'estout.linearmodels_results.r2': ( 'linearmodels_results.html#r2',
                                                                                 'estout/linearmodels_results.py'),
                                             'estout.linearmodels_results.se': ( 'linearmodels_results.html#se',
                                                                                 'estout/linearmodels_results.py'),
                                             'estout.linearmodels_results.tstats': ( 'linearmodels_results.html#tstats',
                                                                                     'estout/linearmodels_results.py'),
                                             'estout.linearmodels_results.xnames': ( 'linearmodels_results.html#xnames',
                                                                                     'estout/linearmodels_results.py'),
                                             'estout.linearmodels_results.ynames': ( 'linearmodels_results.html#ynames',
                                                                                     'estout/linearmodels_results.py')},
            'estout.statsmodels_results': { 'estout.statsmodels_results.covmat': ( 'statsmodels_results.html#covmat',
                                                                                   'estout/statsmodels_results.py'),
                                            'estout.statsmodels_results.nobs': ( 'statsmodels_results.html#nobs',
                                                                                 'estout/statsmodels_results.py'),
                                            'estout.statsmodels_results.params': ( 'statsmodels_results.html#params',
                                                                                   'estout/statsmodels_results.py'),
                                            'estout.statsmodels_results.pvalues': ( 'statsmodels_results.html#pvalues',
                                                                                    'estout/statsmodels_results.py'),
                                            'estout.statsmodels_results.r2': ( 'statsmodels_results.html#r2',
                                                                               'estout/statsmodels_results.py'),
                                            'estout.statsmodels_results.se': ( 'statsmodels_results.html#se',
                                                                               'estout/statsmodels_results.py'),
                                            'estout.statsmodels_results.tstats': ( 'statsmodels_results.html#tstats',
                                                                                   'estout/statsmodels_results.py'),
                                            'estout.statsmodels_results.xnames': ( 'statsmodels_results.html#xnames',
                                                                                   'estout/statsmodels_results.py'),
                                            'estout.statsmodels_results.ynames': ( 'statsmodels_results.html#ynames',
                                                                                   'estout/statsmodels_results.py')},
            'estout.utils': { 'estout.utils.default_formats': ('utils.html#default_formats', 'estout/utils.py'),
                              'estout.utils.df_to_tex': ('utils.html#df_to_tex', 'estout/utils.py'),
                              'estout.utils.get_stars': ('utils.html#get_stars', 'estout/utils.py'),
                              'estout.utils.make_pdf_from_tex': ('utils.html#make_pdf_from_tex', 'estout/utils.py'),
                              'estout.utils.model_groups': ('utils.html#model_groups', 'estout/utils.py'),
                              'estout.utils.open_pdf_file': ('utils.html#open_pdf_file', 'estout/utils.py'),
                              'estout.utils.rgetattr': ('utils.html#rgetattr', 'estout/utils.py'),
                              'estout.utils.rsetattr': ('utils.html#rsetattr', 'estout/utils.py'),
                              'estout.utils.tex_table_env': ('utils.html#tex_table_env', 'estout/utils.py')}}}
